{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\nFormat SeqFISH Data\n===================\n\nThe following script formats SeqFISH data in SpaceTx-Format\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import functools\nimport os\nfrom typing import Mapping, Union\n\nimport click\nimport numpy as np\nimport pandas as pd\nimport skimage.io\nfrom slicedimage import ImageFormat\n\nfrom starfish import Codebook\nfrom starfish.experiment.builder import FetchedTile, TileFetcher, write_experiment_json\nfrom starfish.types import Axes, Coordinates, CoordinateValue, Features\n\n\n# We use this to cache images across tiles.  In the case of the osmFISH data set, volumes are saved\n# together in a single file.  To avoid reopening and decoding the TIFF file, we use a single-element\n# cache that maps between file_path and the npy file.\n@functools.lru_cache(maxsize=1)\ndef cached_read_fn(file_path) -> np.ndarray:\n    return skimage.io.imread(file_path)\n\n\nclass SeqFISHTile(FetchedTile):\n    def __init__(\n            self,\n            file_path: str,\n            coordinates: Mapping[Union[str, Coordinates], CoordinateValue],\n            zplane: int,\n            ch: int,\n    ):\n        self._file_path = file_path\n        self._zplane = zplane\n        self._ch = ch\n        self._coordinates = coordinates\n\n    @property\n    def shape(self) -> Mapping[Axes, int]:\n        \"\"\"\n        Gets image shape directly from the data. Note that this will result in the data being\n        read twice, since the shape is retrieved from all tiles before the data is read, and thus\n        single-file caching does not resolve the duplicated reads.\n\n        Because the data here isn't tremendously large, this is acceptable in this instance.\n        \"\"\"\n        raw_shape = self.tile_data().shape\n        return {Axes.Y: raw_shape[0], Axes.X: raw_shape[1]}\n\n    @property\n    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:\n        \"\"\"Stores coordinate information passed from the TileFetcher\"\"\"\n        return self._coordinates\n\n    @property\n    def format(self) -> ImageFormat:\n        \"\"\"Image Format for SeqFISH data is TIFF\"\"\"\n        return ImageFormat.TIFF\n\n    def tile_data(self) -> np.ndarray:\n        \"\"\"vary z the slowest, then channel -- each round has its own TIFF\"\"\"\n        return cached_read_fn(self._file_path)[self._zplane, self._ch]\n\n\nclass SeqFISHTileFetcher(TileFetcher):\n\n    def __init__(self, input_dir: str) -> None:\n        \"\"\"Implement a TileFetcher for a single SeqFISH Field of View.\"\"\"\n        self.input_dir = input_dir\n\n    @property\n    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:\n        \"\"\"Returns dummy coordinates for this single-FoV TileFetcher\"\"\"\n        return {\n            Coordinates.X: (0., 1.),\n            Coordinates.Y: (0., 1.),\n            Coordinates.Z: (0., 0.1),\n        }\n\n    def get_tile(\n            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> SeqFISHTile:\n        \"\"\"Extracts 2-d data from a multi-page TIFF containing all Tiles for an imaging round\n\n        Parameters\n        ----------\n        fov : int\n            Not used in this implementation, will always receive 0\n        r : int\n            Imaging round. Selects the TIFF file to examine\n        ch : int\n            Selects the channel from within the loaded TIFF file\n        zplane : int\n            Selects the z-plane from within the loaded TIFF file\n\n        Returns\n        -------\n        SeqFISHTile :\n            SeqFISH subclass of FetchedTile\n        \"\"\"\n        file_path = os.path.join(self.input_dir, f\"{round_label + 1}.tif\")\n        return SeqFISHTile(file_path, self.coordinates, zplane_label, ch_label)\n\n\ndef parse_codebook(codebook_csv: str) -> Codebook:\n    \"\"\"Parses a codebook csv file provided by SeqFISH developers.\n\n    Parameters\n    ----------\n    codebook_csv : str\n        The codebook file is expected to contain a matrix whose rows are barcodes and whose columns\n        are imaging rounds. Column IDs are expected to be sequential, and round identifiers (roman\n        numerals) are replaced by integer IDs.\n\n    Returns\n    -------\n    Codebook :\n        Codebook object in SpaceTx format.\n    \"\"\"\n    csv: pd.DataFrame = pd.read_csv(codebook_csv, index_col=0)\n    integer_round_ids = range(csv.shape[1])\n    csv.columns = integer_round_ids\n\n    mappings = []\n\n    for gene, channel_series in csv.iterrows():\n        mappings.append({\n            Features.CODEWORD: [{\n                Axes.ROUND.value: r, Axes.CH.value: c - 1, Features.CODE_VALUE: 1\n            } for r, c in channel_series.items()],\n            Features.TARGET: gene\n        })\n\n    return Codebook.from_code_array(mappings)\n\n\n@click.command()\n@click.option(\"--input-dir\", type=str, required=True, help=\"input directory containing images\")\n@click.option(\"--output-dir\", type=str, required=True, help=\"output directory for formatted data\")\n@click.option(\"--codebook-csv\", type=str, required=True,\n              help=\"csv file containing barcode:target mapping\")\ndef cli(input_dir: str, output_dir: str, codebook_csv: str) -> int:\n    \"\"\"CLI entrypoint for spaceTx format construction for SeqFISH data\n\n    Parameters\n    ----------\n    input_dir : str\n        directory containing input multi-page TIFF files for a single field of view, separated by\n        the imaging round they were acquired in and named <1-index round>.tif\n    output_dir : str\n        directory containing output files. Will be created if it does not exist.\n    codebook_csv : str\n        name of the codebook csv file containing barcode information for this field of view.\n\n    Notes\n    -----\n    - each round is organized as [z, ch, [x|y], [x|y]] -- the order of x and y are not known, but\n      since this script uses dummy coordinates, this distinction is not important.\n    - The spatial organization of the field of view is not known to the starfish developers,\n      so they are filled by dummy coordinates\n    - Raw data (input for this tool) for this experiment can be found at:\n      s3://spacetx.starfish.data.public/browse/raw/seqfish/\n    - Processed data (output of this tool) can be found at:\n      s3://spacetx.starfish.data.public/browse/formatted/20181211/seqfish/ and accessed in\n      `starfish.data.SeqFISH`\n\n    Returns\n    -------\n    int :\n        Returns 0 if successful\n    \"\"\"\n    os.makedirs(output_dir, exist_ok=True)\n    primary_tile_fetcher = SeqFISHTileFetcher(os.path.expanduser(input_dir))\n\n    # This is hardcoded for this example data set\n    primary_image_dimensions: Mapping[Union[str, Axes], int] = {\n        Axes.ROUND: 5,\n        Axes.CH: 12,\n        Axes.ZPLANE: 29,\n    }\n\n    write_experiment_json(\n        path=output_dir,\n        fov_count=1,\n        primary_image_dimensions=primary_image_dimensions,\n        aux_name_to_dimensions={},\n        primary_tile_fetcher=primary_tile_fetcher,\n        tile_format=ImageFormat.TIFF,\n        dimension_order=(Axes.ROUND, Axes.CH, Axes.ZPLANE)\n    )\n\n    # Note: this must trigger AFTER write_experiment_json, as it will clobber the codebook with\n    # a placeholder.\n    codebook = parse_codebook(codebook_csv)\n    codebook.to_json(\"codebook.json\")\n\n    return 0\n\n\nif __name__ == \"__main__\":\n    cli()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.8"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}